//////////////////////////////////////////////////////////////////////////
//	"netspeed.cpp"
//------------------------------------------------------------------------
//	$Id: netspeed.cpp 14 2003-12-13 11:24:38Z taro $
//	2003.12.12	Linux support (Plamo Linux 3.0 / gcc version 2.95.3)
//////////////////////////////////////////////////////////////////////////

#include	"socklib.h"
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>
#if	defined(WIN32)
# include	<process.h>
# include	<conio.h>
#endif
#if defined(LINUX)
# include	<pthread.h>
# include	<sys/time.h>
# include	<unistd.h>
# include	<errno.h>
# include	<ctype.h>
# define	strcmpi	strcasecmp
# define	__stdcall
#endif

class	NetSpeed
{
public:
	enum netspeed_enum
	{
		MODE_RECV = 0,
		MODE_SEND = 1,
		ST_UDP = 0,
		ST_TCP = 1,
		DEFAULT_PORT  = 10240,
		BUFFER_SIZE   = 32768,
		TRANSFER_SIZE = 0x100000,
		MO_WAIT_TIME = 1000,	// WaitForMultipleObjects
		SELECT_WAIT_SEC = 1,
	};

	typedef clock_t			Clock_t;

public:
	int				mode;
	int				stmode;
	char		   *host;
	int				port;
	BOOL			tcpNoDelay;
	unsigned		bufferSize;
	LONGLONG		transferSize;
	int				threadNum;
	int				taskNum;
	struct TaskWork
	{
		NetSpeed		   *ns;
		int					no;
		SockLib				socklib;
		unsigned			bufferSize;
		void			   *buffer;
		LONGLONG			lsize;
		LONGLONG			lsize2;
#if	defined(WIN32)
		unsigned			threadId;
#elif	defined(LINUX)
		BOOL				createThread;
		BOOL				initCond;
		pthread_cond_t		cond;
		BOOL				initMutex;
		pthread_mutex_t		mutex;
		BOOL				runFlag;
		BOOL				runMutex_init;
		pthread_mutex_t		runMutex;
#endif
		SockLib::SockAddr	addr;
		int					errnum;
	};
	TaskWork		*task;
#if		defined(WIN32)
	HANDLE			*htbl;
#elif	defined(LINUX)
	pthread_t		*htbl;
#endif

	BOOL				initCritical;
#if		defined(WIN32)
	CRITICAL_SECTION	critical;
#elif	defined(LINUX)
	pthread_mutex_t		critical;
#endif

	struct Header
	{
		unsigned		size;
		unsigned		seed;
	};

public:
	NetSpeed();

	int		Run(void);

	inline void	GLock(void)
	{
#if		defined(WIN32)
		::EnterCriticalSection(&this->critical);
#elif	defined(LINUX)
		pthread_mutex_lock(&this->critical);
#endif
	}
	inline	void	GUnlock(void)
	{
#if		defined(WIN32)
		::LeaveCriticalSection(&this->critical);
#elif	defined(LINUX)
		pthread_mutex_unlock(&this->critical);
#endif
	}

	Clock_t		Clock(void)
	{
#if		defined(WIN32)
		return (Clock_t)::clock();
#elif	defined(LINUX)
		struct timeval	tv;

		gettimeofday(&tv, NULL);
		return (Clock_t)tv.tv_sec * CLOCKS_PER_SEC + (Clock_t)tv.tv_usec * CLOCKS_PER_SEC / 1000000ll;
#endif
	}

private:
	void	PrintGuide(void);
#if		defined(WIN32)
	static unsigned __stdcall recv_thread(void * param);
	static unsigned __stdcall send_thread(void * param);
#elif	defined(LINUX)
	static void *recv_thread(void * param);
	static void *send_thread(void * param);
#endif
};

#if	defined(LINUX)
static int	kbhit(void)
{
	fd_set			rfd;
	int				n, ret;
	struct timeval	timeout;

	timeout.tv_sec  = 0;
	timeout.tv_usec = 0;
	FD_ZERO(&rfd);
	FD_SET( STDIN_FILENO, &rfd); 
	n = STDIN_FILENO + 1;
	ret = select( n, &rfd, NULL, NULL, &timeout);
	return ret > 0 ? 1 : 0;
}
#endif

NetSpeed::NetSpeed()
{
	this->mode = MODE_RECV;
	this->stmode = ST_UDP;
	this->host = NULL;
	this->port = DEFAULT_PORT;
	this->tcpNoDelay = FALSE;
	this->bufferSize = BUFFER_SIZE;
	this->transferSize = TRANSFER_SIZE;
	this->threadNum = 1;
	this->taskNum = 0;
	this->task = NULL;
	this->htbl = NULL;

	this->initCritical = FALSE;
}


int		NetSpeed::Run(void)
{
	int				i;
	Clock_t			clk0, clk1;
	int				retcode;

	if ( (this->task = new TaskWork[this->threadNum]) == NULL )
		return -1;
	this->taskNum = this->threadNum;
	for ( i = 0; i < this->taskNum; i++ )
	{
		this->task[i].ns = this;
		this->task[i].no = i;
		this->task[i].buffer = NULL;
		this->task[i].lsize = 0;
		this->task[i].lsize2 = (LONGLONG)-1;
		this->task[i].errnum = 0;
#if	defined(LINUX)
		this->task[i].createThread  = FALSE;
		this->task[i].initCond      = FALSE;
		this->task[i].initMutex     = FALSE;
		this->task[i].runFlag       = FALSE;
		this->task[i].runMutex_init = FALSE;
#endif
	}

	try
	{
#if		defined(WIN32)
		if ( (this->htbl = new HANDLE[this->threadNum]) == NULL )
			throw -1;
		for ( i = 0; i < this->taskNum; i++ )
			this->htbl[i] = NULL;
		::InitializeCriticalSection(&this->critical);
#elif defined(LINUX)
		if ( (this->htbl = new pthread_t[this->threadNum]) == NULL )
			throw -1;
		pthread_mutex_init(&this->critical,NULL);
#endif
		this->initCritical = TRUE;

		for ( i = 0; i < this->taskNum; i++ )
		{
			int					socktype = SOCK_DGRAM;
			TaskWork		   *task;
			const char		   *protname;

			fprintf(stderr,"thread %d : ", i + 1);
			task = &this->task[i];
			switch ( this->stmode )
			{
				case ST_UDP: socktype = SOCK_DGRAM ; protname = "UDP"; break;
				case ST_TCP: socktype = SOCK_STREAM; protname = "TCP"; break;
			}
			if ( task->socklib.Create(socktype) < 0 )
			{
				fprintf(stderr," socklib create error\n");
				throw -1;
			}
			task->bufferSize = this->bufferSize;
#if		defined(WIN32)
			if ( this->stmode == ST_UDP )
			{
				unsigned int	optval;
				int				optsize;

				optval = 0;
				optsize = sizeof(optval);
				if ( task->socklib.GetSockOpt(SOL_SOCKET, SO_MAX_MSG_SIZE, &optval, &optsize) < 0 )
				{
					fprintf(stderr," socklib getsockopt error\n");
					throw -1;
				}
				if ( optval > 0 && this->bufferSize > optval )
				{
					task->bufferSize = optval;
					if ( i == 0 )
						fprintf(stderr," change buffer size %u : ", task->bufferSize);
				}
			}
#endif
			if ( (task->buffer = new BYTE [task->bufferSize]) == NULL )
				throw -1;
			memset( task->buffer, 0, task->bufferSize);
			if ( this->mode == MODE_SEND )
			{
				unsigned	seed;
				Header	   *head;
				BYTE	   *p;

				seed = (unsigned)clock();
				srand(seed);
				head = (Header *)task->buffer;
				head->size = task->bufferSize;
				head->seed = seed;
				for ( p = (BYTE *)(head + 1); p < (BYTE *)task->buffer + task->bufferSize; p++ )
					*p = rand();
			}

			if ( this->stmode == ST_TCP )
			{
				BOOL	flag;

				flag = this->tcpNoDelay ? 1 : 0;
				if ( task->socklib.SetSockOpt(IPPROTO_TCP, TCP_NODELAY, &flag, sizeof(flag)) < 0 )
				{
					fprintf(stderr," socklib setsockopt(tcp_nodelay) error\n");
					throw -1;
				}
			}

			switch ( this->mode )
			{
				case MODE_RECV:
					task->addr.sockaddr.sin_family = AF_INET;
					task->addr.sockaddr.sin_addr.s_addr = INADDR_ANY;
					task->addr.SetPort(this->port + i);
					fprintf(stderr," bind any (port %u) %s", port + i, protname);
					if ( task->socklib.Bind(task->addr) < 0 )
					{
						fprintf(stderr," socklib bind error\n");
						throw -1;
					}
					switch ( this->stmode )
					{
						case ST_UDP:
							break;
						case ST_TCP:
							task->socklib.Listen(1);
							break;
					}
					break;

				case MODE_SEND:
				{
					task->addr.sockaddr.sin_family = AF_INET;
					task->addr.SetAddr(this->host);
					task->addr.SetPort(this->port + i);
					fprintf(stderr," connect %s (port %u) %s", inet_ntoa(task->addr.sockaddr.sin_addr), port + i, protname);
					switch ( this->stmode )
					{
						case ST_UDP:
							break;
						case ST_TCP:
							if ( task->socklib.Connect(task->addr) < 0 )
							{
								fprintf(stderr," socklib connect error\n");
								throw -1;
							}
							break;
					}
					break;
				}
			}
			fprintf(stderr,"\n");
		}

		this->GLock();
		switch ( this->mode )
		{
			case MODE_RECV:
				for ( i = 0; i < this->taskNum; i++ )
				{
#if		defined(WIN32)
					if ( (this->htbl[i] = (HANDLE)_beginthreadex( NULL, 0, recv_thread, &this->task[i], 0, &this->task[i].threadId)) == 0 )
						throw -1;
#elif	defined(LINUX)
					pthread_mutex_init(&this->task[i].runMutex, NULL);
					this->task[i].runMutex_init = TRUE;
					pthread_mutex_lock(&this->task[i].runMutex);
					this->task[i].runFlag = TRUE;
					pthread_mutex_unlock(&this->task[i].runMutex);

					pthread_mutex_init(&this->task[i].mutex, NULL);
					this->task[i].initMutex = TRUE;

					if ( pthread_create( &this->htbl[i], NULL, recv_thread, &this->task[i]) != 0 )
						throw -1;
					this->task[i].createThread = TRUE;
#endif
				}
				break;

			case MODE_SEND:
				for ( i = 0; i < this->taskNum; i++ )
				{
#if		defined(WIN32)
					if ( (this->htbl[i] = (HANDLE)_beginthreadex( NULL, 0, send_thread,  &this->task[i], 0, &this->task[i].threadId)) == 0 )
						throw -1;
#elif	defined(LINUX)
					pthread_mutex_init(&this->task[i].runMutex, NULL);
					this->task[i].runMutex_init = TRUE;
					pthread_mutex_lock(&this->task[i].runMutex);
					this->task[i].runFlag = TRUE;
					pthread_mutex_unlock(&this->task[i].runMutex);

					pthread_mutex_init(&this->task[i].mutex, NULL);
					this->task[i].initMutex = TRUE;

					if ( pthread_create( &this->htbl[i], NULL, send_thread, &this->task[i]) != 0 )
						throw -1;
					this->task[i].createThread = TRUE;
#endif
				}
				break;
		}
		clk0 = this->Clock();
		this->GUnlock();
		for (;;)
		{
#if		defined(WIN32)
			DWORD	result;

			result = ::WaitForMultipleObjects(this->taskNum, this->htbl, TRUE, NetSpeed::MO_WAIT_TIME);
			if ( result != WAIT_TIMEOUT || _kbhit() != 0 )
				break;
#elif	defined(LINUX)
			struct timeval	tv;
			struct timespec	ts;
			BOOL			endflag;

			if ( kbhit() != 0 )
				break;
			gettimeofday(&tv, NULL);
			ts.tv_sec  = tv.tv_sec  +  NetSpeed::MO_WAIT_TIME / 1000;
			ts.tv_nsec = tv.tv_usec + (NetSpeed::MO_WAIT_TIME % 1000);
			endflag = TRUE;
			for ( i = 0; i < this->taskNum; i++ )
			{
				int		res;

				pthread_mutex_lock(&this->task[i].runMutex);
				res = this->task[i].runFlag;
				pthread_mutex_unlock(&this->task[i].runMutex);
				if ( res )
				{
					pthread_mutex_lock(&this->task[i].mutex);
					res = pthread_cond_timedwait(&this->task[i].cond, &this->task[i].mutex, &ts);
					if ( res == ETIMEDOUT )
						endflag = FALSE;
					if ( res == EINTR )
						pthread_join(this->htbl[i], NULL);
					pthread_mutex_unlock(&this->task[i].mutex);
					if ( res == ETIMEDOUT )
						break;
				}
			}
			if ( endflag )
				break;
#endif
			this->PrintGuide();
		}
		clk1 = this->Clock();
		for ( i = 0; i < this->taskNum; i++ )
		{
#if		defined(WIN32)
			::CloseHandle(this->htbl[i]);
#elif	defined(LINUX)
			int		res;

			pthread_mutex_lock(&this->task[i].runMutex);
			res = this->task[i].runFlag;
			pthread_mutex_unlock(&this->task[i].runMutex);
			if ( res )
				pthread_detach(this->htbl[i]);
#endif
		}

		this->PrintGuide();
		switch ( this->mode )
		{
			case MODE_RECV:
			{
				LONGLONG	size;

				size = 0;
				for ( i = 0; i < this->taskNum; i++ )
					size += this->task[i].lsize;
				fprintf(stdout, "recv %u K bytes\n", (unsigned)(size/(LONGLONG)0x400));
				break;
			}

			case MODE_SEND:
				if ( clk1 > clk0 )
				{
					LONGLONG	size;
					unsigned	speed;
					Clock_t		clk;

					clk = clk1 - clk0;
					size = 0;
					for ( i = 0; i < this->taskNum; i++ )
						size += this->task[i].lsize;
					speed = (unsigned)(size * (LONGLONG)CLOCKS_PER_SEC / (LONGLONG)clk);
					fprintf(stdout, "send %u K bytes, speed %u bytes/sec (%u clocks)\n", 
					  (unsigned)(size/(LONGLONG)0x400),
					  speed,
					  clk);
				}
				break;
		}
		retcode = 0;
	} // end of try
	catch (int code) { retcode = code; }

	if ( this->initCritical )
	{
#if		defined(WIN32)
		::DeleteCriticalSection(&this->critical);
#elif	defined(LINUX)
		pthread_mutex_destroy(&this->critical);
#endif
		this->initCritical = FALSE;
	}
	if ( this->htbl != NULL )
	{
		delete [] this->htbl;
		this->htbl = NULL;
	}
	if ( this->task != NULL )
	{
		for ( i = 0; i < this->taskNum; i++ )
		{
			if ( this->task[i].buffer != NULL )
				delete [] (BYTE *)this->task[i].buffer;
		}
		delete [] this->task;
		this->task = NULL;
	}
	return retcode;
}

void	NetSpeed::PrintGuide(void)
{
	BOOL	disp;
	int		i;

	this->GLock();
	disp = FALSE;
	for ( i = 0; i < this->taskNum; i++ )
	{
		if ( this->task[i].lsize != this->task[i].lsize2 )
		{
			disp = TRUE;
			break;
		}
	}
	if ( disp )
	{
		for ( i = 0; i < this->taskNum; i++ )
		{
			fprintf(stderr, " %12u ", (unsigned)(this->task[i].lsize/(LONGLONG)0x400) );
			this->task[i].lsize2 = this->task[i].lsize;
		}
		fprintf(stderr, "\n");
	}
	this->GUnlock();
}

#if		defined(WIN32)
unsigned __stdcall NetSpeed::recv_thread(void * param)
#elif	defined(LINUX)
void *	NetSpeed::recv_thread(void * param)
#endif
{
	NetSpeed::TaskWork	*task = (NetSpeed::TaskWork *)param;
	NetSpeed			*ns = task->ns;
	SockLib::SockAddr	from;
	SockLib				msgsock;
	int					ret;

	ns->GLock();
	fprintf(stderr,"start thread %d\n", task->no + 1);
	ns->GUnlock();

	while ( task->lsize < ns->transferSize )
	{
		switch ( ns->stmode )
		{
			case ST_UDP:
			{
				fd_set			fds;
				SOCKET			fd;
				struct timeval	tm;

				tm.tv_sec  = NetSpeed::SELECT_WAIT_SEC;
				tm.tv_usec = 0;
				fd = task->socklib.GetSocket();
				FD_ZERO(&fds);
				FD_SET(fd, &fds);
				::select(fd + 1, &fds, NULL, NULL, &tm);	// wait recv
				if ( FD_ISSET(fd, &fds) )
					ret = task->socklib.RecvFrom(task->bufferSize, task->buffer, 0, from);
				else
					ret = 0;
				break;
			}

			case ST_TCP:
				if ( msgsock.IsCreate() == FALSE )
				{
					if ( task->socklib.Accept(&msgsock, from) < 0 )
						goto END;
				}
				ret = msgsock.Recv(task->bufferSize, task->buffer, 0);
				if ( ret <= 0 )
					msgsock.Close();
				break;
		}
		if ( (ret == 0 && task->lsize > 0) )
			break;
		if ( ret > 0 )
			task->lsize += ret;
	}

END:
#if		defined(WIN32)
	 _endthreadex(0);
	return 0;
#elif	defined(LINUX)
	pthread_mutex_lock(&(task->runMutex));
	task->runFlag = FALSE;
	pthread_mutex_unlock(&(task->runMutex));
	pthread_cond_signal(&(task->cond));
	pthread_exit(NULL);
	return NULL;
#endif
}

#if	defined(WIN32)
unsigned __stdcall NetSpeed::send_thread(void * param)
#elif	defined(LINUX)
void *	NetSpeed::send_thread(void * param)
#endif
{
	NetSpeed::TaskWork	*task = (NetSpeed::TaskWork *)param;
	NetSpeed			*ns = task->ns;
	int					ret;

	ns->GLock();
	fprintf(stderr,"start thread %d\n", task->no + 1);
	ns->GUnlock();

	while ( task->lsize < ns->transferSize )
	{
		switch ( ns->stmode )
		{
			case ST_UDP:
				ret = task->socklib.SendTo(task->bufferSize, task->buffer,0, task->addr);
				break;

			case ST_TCP:
				ret = task->socklib.Send(task->bufferSize, task->buffer,0);
				break;
		}
		if ( ret > 0 )
		{
		//	ns->GLock();
		//	fprintf(stderr,"send %u bytes\n", ret);
		//	ns->GUnlock();
			task->lsize += ret;
		}
	}

#if		defined(WIN32)
	 _endthreadex(0);
#elif	defined(LINUX)
	pthread_mutex_lock(&(task->runMutex));
	task->runFlag = FALSE;
	pthread_mutex_unlock(&(task->runMutex));
	pthread_cond_signal(&(task->cond));
	pthread_exit(NULL);
#endif
	return 0;
}


void	help(void)
{
	static const char *msg[] =
	{
		"usage: netspeed <option>\n",
		"   -recv              ... receive(server) mode (default)\n",
		"   -send <host/addr>  ... send(client) mode\n",
		"   -udp               ... UDP (default)\n",
		"   -tcp               ... TCP\n",
		"   -tcpnodelay        ... TCP NODELAY with -tcp\n",
		"   -port <n>          ... port number\n",
		"   -thread <n>        ... thread number\n",
		"   -size <size>       ... transfer size (<n> + G, M, K)\n",
		"   -buffer <size>     ... buffer(packet) size\n",
	};
	int				i;

	for ( i = 0; i < sizeof(msg)/sizeof(*msg); i++ )
		fputs(msg[i], stdout);
}

int		main(int argc, char **argv)
{
	int				i;
	NetSpeed		ns;

	for ( i = 1; i < argc; i++ )
	{
		const char	*s;

		s = argv[i];
		if ( s[0] == '-' )
		{	// option
			++s;
			if ( strcmpi(s, "help") == 0 || s[0] == '?' )
			{
				help();
				return 1;
			} else if ( strcmpi(s, "recv") == 0 )
			{
				ns.mode = NetSpeed::MODE_RECV;
			} else if ( strcmpi(s, "send") == 0 )
			{
				ns.mode = NetSpeed::MODE_SEND;
				if ( i + 1 < argc && argv[i+1][0] != '-' )
				{	// host
					s = argv[++i];
					ns.host = (char *)s;
				}
			} else if ( strcmpi(s, "udp") == 0 )
			{
				ns.stmode = NetSpeed::ST_UDP;
			} else if ( strcmpi(s, "tcp") == 0 )
			{
				ns.stmode = NetSpeed::ST_TCP;
			} else if ( strcmpi(s, "tcpnodelay") == 0 )
			{
				ns.tcpNoDelay = TRUE;
			} else if ( strcmpi(s, "port") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				struct protoent	   *pe;

				s = argv[++i];
				if ( isdigit(s[0]) )
				{
					ns.port = atoi(s);
				} else
				{
					if ( (pe = ::getprotobyname(s)) != NULL )
						ns.port = pe->p_proto;
				}
			} else if ( strcmpi(s, "thread") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				int			n;

				s = argv[++i];
				n = atoi(s);
				if ( n > 1 )
					ns.threadNum = n;
			} else if ( strcmpi(s, "size") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				int			n, len;

				s = argv[++i];
				len = strlen(s);
				n = atoi(s);
				if ( n > 1 )
				{
					switch ( s[len-1] )
					{
						case 'G':	// giga
						case 'g':
							ns.transferSize = (LONGLONG)n * (LONGLONG)0x40000000;
							break;

						case 'M':	// mega
						case 'm':
							ns.transferSize = (LONGLONG)n * (LONGLONG)0x100000;
							break;

						default:
						case 'K':	// kilo
						case 'k':
							ns.transferSize = (LONGLONG)n * (LONGLONG)0x400;
							break;
					}
				}
			} else if ( strcmpi(s, "buffer") == 0 && i + 1 < argc && argv[i+1][0] != '-' )
			{
				int			n;

				s = argv[++i];
				n = atoi(s);
				if ( n > 1 && n <= 0x10000 )
					ns.bufferSize = n;
			} else
			{
				fprintf(stderr, "parameter error %s\n", argv[i]);
				return 1;
			}
		} else
		{
			fprintf(stderr, "parameter error %s\n", argv[i]);
			return 1;
		}
	}
	if ( (LONGLONG)ns.bufferSize > ns.transferSize )
		ns.bufferSize = (unsigned)ns.transferSize;
	switch ( ns.mode )
	{
		case NetSpeed::MODE_RECV:
			fprintf(stderr,"Recv mode\n");
			break;

		case NetSpeed::MODE_SEND:
			fprintf(stderr,"Send mode\n");
			if ( ns.host == NULL || ns.host[0] == '\0' )
			{
				fprintf(stderr,"host error!\n");
				return 1;
			}
			break;
	}
	fprintf(stderr, "thread %d, buffer size %u bytes, transfer size %u K bytes\n",
	  ns.threadNum,
	  ns.bufferSize,
	  (unsigned)(ns.transferSize/(LONGLONG)0x400) );
	if ( ns.Run() < 0 )
		return 1;

	return 0;
}
