//////////////////////////////////////////////////////////////////////////
//	"socklib.cpp"
//------------------------------------------------------------------------
//	$Id: socklib.cpp 15 2003-12-13 11:26:16Z taro $
//////////////////////////////////////////////////////////////////////////

#include	"socklib.h"
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

//////////////////////////////////////////////////////////////////////////
//	SockAddr
//////////////////////////////////////////////////////////////////////////

SockLib::SockAddr::SockAddr()
{
	::memset(&this->sockaddr, 0, sizeof(this->sockaddr));
	this->sockaddr.sin_family = AF_INET;
}

int		SockLib::SockAddr::SetAddr(unsigned long addr, unsigned port)
{
	memset(&this->sockaddr, 0, sizeof(this->sockaddr));
	this->sockaddr.sin_family = AF_INET;
	this->SetAddr(addr);
	this->SetPort(port);
	return 0;
}

int		SockLib::SockAddr::SetAddr(const char * s)
{
	struct hostent	*hp;

	if ( isalpha(s[0]) )
	{	// name
		hp = ::gethostbyname(s);
	} else
	{	// xxx.xxx.xxx.xxx
		unsigned long	a;

		if ( (a = ::inet_addr(s)) == INADDR_NONE )
			return -1;
		hp = ::gethostbyaddr((char *)&a, 4, AF_INET);
	}
	if ( hp == NULL )
		return -1;
	this->sockaddr.sin_family = hp->h_addrtype;
	memcpy(&this->sockaddr.sin_addr, hp->h_addr, hp->h_length);
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	SockLib
//////////////////////////////////////////////////////////////////////////

int			SockLib::sockCount  = 0;
#if defined(WIN32)
BOOL		SockLib::wsaStartup = FALSE;
WORD		SockLib::sockVer    = MAKEWORD(2, 0);
WSADATA		SockLib::wsaData;
#endif

SockLib::SockLib()
{
	this->flags = 0;
	this->socket_type = SOCK_STREAM;	// tcp
	this->conn_socket = 0;
}

SockLib::~SockLib()
{
	if ( this->flags )
		this->Close();
}

int		SockLib::Create(int type)
{
	if ( this->flags & F_SOCKET )
		return -1;

#if defined(WIN32)
	if ( SockLib::wsaStartup == FALSE )
	{
		if ( ::WSAStartup(SockLib::sockVer, &SockLib::wsaData) != 0 )
		{
			::WSACleanup();
			return -1;
		}
		SockLib::wsaStartup = TRUE;
	}
#endif

	this->socket_type = type;
	this->conn_socket = ::socket(AF_INET, this->socket_type, 0);
	if ( this->conn_socket < 0 )
	{
		this->Close();
		return -1;	// error
	}
	this->flags |= F_SOCKET;
	SockLib::sockCount++;
	return 0;
}

int		SockLib::Close(void)
{
	if ( this->flags & F_SOCKET )
	{
#if	defined(WIN32)
		::closesocket(this->conn_socket);
#else
		close(this->conn_socket);
#endif
		this->flags &= ~F_SOCKET;
		SockLib::sockCount--;
	}
#if	defined(WIN32)
	if ( SockLib::sockCount == 0 && SockLib::wsaStartup != FALSE )
	{
		::WSACleanup();
		SockLib::wsaStartup = FALSE;
	}
#endif
	return 0;
}

int		SockLib::SetSocket(SOCKET socket)
{
	if ( this->flags & F_SOCKET )
		return -1;
	this->conn_socket = socket;
	this->flags |= F_SOCKET;
	SockLib::sockCount++;
	return 0;
}

int		SockLib::GetSockOpt(int level, int optname, void * optval, socklen_t * optlen)
{
	int			ret;

	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( (ret = ::getsockopt(this->conn_socket, level, optname, (char *)optval, optlen)) == SOCKET_ERROR )
		return -1;
	return ret;
}

int		SockLib::SetSockOpt(int level, int optname, void * optval, socklen_t optlen)
{
	int			ret;

	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( (ret = ::setsockopt(this->conn_socket, level, optname, (char *)optval, optlen)) != 0 )
		return ret;
	return ret;
}

int		SockLib::Connect(SockLib::SockAddr & addr)
{
	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( ::connect(this->conn_socket, (struct sockaddr *)&addr.sockaddr, sizeof(addr.sockaddr)) == SOCKET_ERROR )
		return -1;
	return 0;
}

int		SockLib::Bind(SockLib::SockAddr & addr)
{
	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( ::bind(this->conn_socket, (struct sockaddr *)&addr.sockaddr, sizeof(addr.sockaddr)) == SOCKET_ERROR )
		return -1;
	return 0;
}

int		SockLib::Listen(int backlog)
{
	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( ::listen(this->conn_socket, backlog) == SOCKET_ERROR )
		return -1;
	return 0;
}

int		SockLib::Accept(SockLib * socklib, SockLib::SockAddr & addr)
{
	socklen_t	addrlen = sizeof(addr.sockaddr);
	SOCKET		socket;

	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( (socket = ::accept(this->conn_socket, (struct sockaddr *)&addr.sockaddr, &addrlen)) == SOCKET_ERROR )
		return -1;
	if ( socklib != NULL )
		socklib->SetSocket(socket);
	return 0;
}

int		SockLib::Send(unsigned size, const void * buffer, int flag)
{
	int		ret;

	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( (ret = ::send(this->conn_socket, (const char *)buffer, size, flag)) == SOCKET_ERROR )
		return -1;
	return ret;
}

int		SockLib::SendTo(unsigned size, const void * buffer, int flag, SockLib::SockAddr & toaddr)
{
	int		ret;

	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( (ret = ::sendto(this->conn_socket, (const char *)buffer, size, flag, (struct sockaddr *)&toaddr.sockaddr, sizeof(toaddr.sockaddr))) == SOCKET_ERROR )
		return -1;
	return ret;
}

int		SockLib::Recv(unsigned size, void * buffer, int flag)
{
	int		ret;

	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( (ret = ::recv(this->conn_socket, (char *)buffer, size, flag)) == SOCKET_ERROR )
		return -1;
	return ret;
}

int		SockLib::RecvFrom(unsigned size, void * buffer, int flag, SockLib::SockAddr & from)
{
	int			ret;
	socklen_t	fromlen = sizeof(from.sockaddr);

	if ( (this->flags & F_SOCKET) == 0 )
		return -1;
	if ( (ret = ::recvfrom(this->conn_socket, (char *)buffer, size, flag, (struct sockaddr *)&from.sockaddr, &fromlen)) == SOCKET_ERROR )
		return -1;
	return ret;
}
